from active_bo_msgs.msg import ActiveBORequest
from active_bo_msgs.msg import ActiveBOResponse
from active_bo_msgs.msg import ActiveRL
from active_bo_msgs.msg import ActiveRLResponse

import rclpy
from rclpy.node import Node

from rclpy.callback_groups import ReentrantCallbackGroup

from active_bo_ros.BayesianOptimization.BayesianOptimization import BayesianOptimization
from active_bo_ros.ReinforcementLearning.ContinuousMountainCar import Continuous_MountainCarEnv

import numpy as np
import time


class ActiveBOTopic(Node):
    def __init__(self):
        super().__init__('active_bo_topic')

        bo_callback_group = ReentrantCallbackGroup()
        rl_callback_group = ReentrantCallbackGroup()
        mainloop_callback_group = ReentrantCallbackGroup()

        # Active Bayesian Optimization Publisher, Subscriber and Message attributes
        self.active_bo_pub = self.create_publisher(ActiveBOResponse,
                                                   'active_bo_response',
                                                   1, callback_group=bo_callback_group)

        self.active_bo_sub = self.create_subscription(ActiveBORequest,
                                                      'active_bo_request',
                                                      self.active_bo_callback,
                                                      1, callback_group=bo_callback_group)

        self.active_bo_pending = False
        self.bo_nr_weights = None
        self.bo_steps = None
        self.bo_episodes = None
        self.bo_runs = None
        self.bo_acq_fcn = None
        self.bo_epsilon = None
        self.current_run = 0
        self.current_episode = 0

        # Active Reinforcement Learning Publisher, Subscriber and Message attributes
        self.active_rl_pub = self.create_publisher(ActiveRL,
                                                   'active_rl_request',
                                                   1, callback_group=rl_callback_group)
        self.active_rl_sub = self.create_subscription(ActiveRLResponse,
                                                      'active_rl_response',
                                                      self.active_rl_callback,
                                                      1, callback_group=rl_callback_group)

        self.active_rl_pending = False
        self.rl_weights = None
        self.rl_final_step = None
        self.rl_reward = None

        # RL Environments and BO
        self.env = Continuous_MountainCarEnv()
        self.distance_penalty = 0

        self.BO = None
        self.nr_init = 3
        self.reward = None
        self.best_pol_reward = None
        self.best_policy = None
        self.best_weights = None

        # Main loop timer object
        self.mainloop_timer_period = 0.1
        self.mainloop = self.create_timer(self.mainloop_timer_period,
                                          self.mainloop_callback,
                                          callback_group=mainloop_callback_group)

    def reset_bo_request(self):
        self.bo_nr_weights = None
        self.bo_steps = None
        self.bo_episodes = None
        self.bo_runs = None
        self.bo_acq_fcn = None
        self.bo_epsilon = None
        self.current_run = 0
        self.current_episode = 0

    def active_bo_callback(self, msg):
        if not self.active_bo_pending:
            self.get_logger().info('Active Bayesian Optimization request pending!')
            self.active_bo_pending = True
            self.bo_nr_weights = msg.nr_weights
            self.bo_steps = msg.max_steps
            self.bo_episodes = msg.nr_episodes
            self.bo_runs = msg.nr_runs
            self.bo_acq_fcn = msg.acquisition_function
            self.bo_epsilon = msg.epsilon

            # initialize
            self.reward = np.zeros((self.bo_episodes, self.bo_runs))
            self.best_pol_reward = np.zeros((1, self.bo_runs))
            self.best_policy = np.zeros((self.bo_steps, self.bo_runs))
            self.best_weights = np.zeros((self.bo_nr_weights, self.bo_runs))

    def reset_rl_response(self):
        self.rl_weights = None
        self.rl_final_step = None
        self.rl_reward = None

    def active_rl_callback(self, msg):
        if self.active_rl_pending:
            self.get_logger().info('Active Reinforcement Learning response pending!')
            self.active_rl_pending = False
            self.rl_weights = msg.weights
            self.rl_final_step = msg.final_step
            self.rl_reward = msg.reward

    def mainloop_callback(self):
        if self.active_bo_pending:
            if self.BO is None:
                self.BO = BayesianOptimization(self.env,
                                               self.bo_steps,
                                               nr_init=self.nr_init,
                                               acq=self.bo_acq_fcn,
                                               nr_weights=self.bo_nr_weights)

                self.BO.initialize()

            if self.current_run == self.bo_runs:
                bo_response = ActiveBOResponse()

                best_policy_idx = np.argmax(self.best_pol_reward)
                bo_response.best_policy = self.best_policy[:, best_policy_idx].tolist()
                bo_response.best_weights = self.best_weights[:, best_policy_idx].tolist()

                bo_response.reward_mean = np.mean(self.reward, axis=1).tolist()
                bo_response.reward_std = np.std(self.reward, axis=1).tolist()

                self.get_logger().info('Responding: Active BO')
                self.active_bo_pub.publish(bo_response)
                self.reset_bo_request()
                self.active_bo_pending = False

            else:
                if self.active_rl_pending:
                    pass
                elif self.rl_weights is not None and not self.active_rl_pending:
                    try:
                        self.BO.add_new_observation(self.rl_reward, self.rl_weights)
                        self.reset_rl_response()
                    except Exception as e:
                        self.get_logger().error(f'Active Reinforcement Learning failed to add new observation: {e}')
                else:
                    if self.current_episode < self.bo_episodes:
                        if np.random.uniform(0.0, 1.0, 1) < self.bo_epsilon:
                            active_rl_request = ActiveRL()
                            old_policy, _, old_weights = self.BO.get_best_result()

                            active_rl_request.policy = old_policy.tolist()
                            active_rl_request.weights = old_weights.tolist()

                            self.get_logger().info('Calling: Active RL')
                            self.active_rl_pub.publish(active_rl_request)
                            self.active_rl_pending = True

                        else:
                            x_next = self.BO.next_observation()
                            self.BO.eval_new_observation(x_next)

                        self.current_episode += 1
                        self.get_logger().info(f'Current Episode: {self.current_episode}')
                    else:
                        self.best_policy[:, self.current_run], \
                            self.best_pol_reward[:, self.current_run], \
                            self.best_weights[:, self.current_run] = self.BO.get_best_result()

                        self.reward[:, self.current_run] = self.BO.best_reward.T

                        self.BO = None

                        self.current_episode = 0
                        self.current_run += 1
                        self.get_logger().info(f'Current Run: {self.current_run}')



def main(args=None):
    rclpy.init(args=args)

    active_bo_topic = ActiveBOTopic()

    rclpy.spin(active_bo_topic)

    try:
        rclpy.spin(active_bo_topic)
    except KeyboardInterrupt:
        pass

    active_bo_topic.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
