from active_bo_msgs.msg import ActiveRL
from active_bo_msgs.msg import ActiveRLResponse

from active_bo_msgs.msg import ImageFeedback

import rclpy
from rclpy.node import Node

from rclpy.callback_groups import ReentrantCallbackGroup

from active_bo_ros.ReinforcementLearning.ContinuousMountainCar import Continuous_MountainCarEnv

import numpy as np
import time
import copy


class ActiveRLService(Node):
    def __init__(self):
        super().__init__('active_rl_service')
        rl_callback_group = ReentrantCallbackGroup()
        topic_callback_group = ReentrantCallbackGroup()
        mainloop_callback_group = ReentrantCallbackGroup()

        # Active Reinforcement Learning Publisher, Subscriber and Message attributes
        self.active_rl_pub = self.create_publisher(ActiveRLResponse,
                                                   'active_rl_response',
                                                   1, callback_group=rl_callback_group)
        self.active_rl_sub = self.create_subscription(ActiveRL,
                                                      'active_rl_request',
                                                      self.active_rl_callback,
                                                      1, callback_group=rl_callback_group)

        self.active_rl_pending = False
        self.rl_policy = None
        self.rl_weights = None
        self.rl_reward = 0.0
        self.rl_step = 0

        # Image publisher to publish the rgb array from the gym environment
        self.image_pub = self.create_publisher(ImageFeedback,
                                               'rl_feedback',
                                               1, callback_group=topic_callback_group)

        # Active RL Evaluation Publisher, Subscriber and Message attributes
        self.eval_pub = self.create_publisher(ActiveRL,
                                              'active_rl_eval_request',
                                              1,
                                              callback_group=topic_callback_group)
        self.eval_sub = self.create_subscription(ActiveRL,
                                                 'active_rl_eval_response',
                                                 self.active_rl_eval_callback,
                                                 1,
                                                 callback_group=topic_callback_group)

        self.eval_response_received = False
        self.eval_policy = None
        self.eval_weights = None

        # RL Environments
        self.env = Continuous_MountainCarEnv(render_mode='rgb_array')
        self.distance_penalty = 0
        self.best_pol_shown = False

        # Main loop timer object
        self.mainloop_timer_period = 0.05
        self.mainloop = self.create_timer(self.mainloop_timer_period,
                                          self.mainloop_callback,
                                          callback_group=mainloop_callback_group)

    def reset_rl_request(self):
        self.rl_policy = None
        self.rl_weights = None

    def active_rl_callback(self, msg):
        self.rl_policy = msg.policy
        self.rl_weights = msg.weights

        self.get_logger().info('Active RL: Called!')
        self.env.reset()
        self.active_rl_pending = True

    def reset_eval_request(self):
        self.eval_policy = None
        self.eval_weights = None

    def active_rl_eval_callback(self, msg):
        self.eval_policy = msg.policy
        self.eval_weights = msg.weights

        self.get_logger().info('Active RL Eval: Responded!')
        self.eval_response_received = True

    def next_image(self, policy):
        action = policy[self.rl_step]
        output = self.env.step(action)

        self.rl_reward += output[1]
        done = output[2]
        self.rl_step += 1

        rgb_array = self.env.render()
        rgb_shape = rgb_array.shape

        red = rgb_array[:, :, 0].flatten().tolist()
        green = rgb_array[:, :, 1].flatten().tolist()
        blue = rgb_array[:, :, 2].flatten().tolist()

        feedback_msg = ImageFeedback()

        feedback_msg.height = rgb_shape[0]
        feedback_msg.width = rgb_shape[1]
        feedback_msg.red = red
        feedback_msg.green = green
        feedback_msg.blue = blue

        self.image_pub.publish(feedback_msg)

        if not done and self.rl_step == len(policy):
            distance = -(self.env.goal_position - output[0][0])
            self.rl_reward += distance * self.distance_penalty
            done = True

        return done

    def mainloop_callback(self):
        if self.active_rl_pending:
            if not self.best_pol_shown:
                done = self.next_image(self.rl_policy)

                if done:
                    self.rl_step = 0
                    self.rl_reward = 0.0

                    eval_request = ActiveRL()
                    eval_request.policy = self.rl_policy
                    eval_request.weights = self.rl_weights

                    self.eval_pub.publish(eval_request)
                    self.get_logger().info('Active RL: Called!')
                    self.get_logger().info('Active RL: Waiting for Eval!')

                    self.env.reset()

                    self.best_pol_shown = True

            elif self.best_pol_shown:
                if not self.eval_response_received:
                    pass

                if self.eval_response_received:
                    done = self.next_image(self.eval_policy)

                    if done:
                        rl_response = ActiveRLResponse()
                        rl_response.weights = self.rl_weights
                        rl_response.reward = self.rl_reward
                        rl_response.final_step = self.rl_step

                        self.active_rl_pub.publish(rl_response)

                        self.env.reset()

                        # reset flags and attributes
                        self.reset_eval_request()
                        self.reset_rl_request()

                        self.rl_step = 0
                        self.rl_reward = 0.0

                        self.best_pol_shown = False
                        self.eval_response_received = False
                        self.active_rl_pending = False


def main(args=None):
    rclpy.init(args=args)

    active_rl_service = ActiveRLService()

    rclpy.spin(active_rl_service)

    rclpy.shutdown()


if __name__ == '__main__':
    main()
